import logging
import json
import sys
import time
import requests
import asyncio

from aliyunsdkcore.client import AcsClient
from aliyunsdkslb.request.v20140515.AddBackendServersRequest import AddBackendServersRequest
from aliyunsdkslb.request.v20140515.RemoveBackendServersRequest import RemoveBackendServersRequest
from aliyunsdkecs.request.v20140526.RunInstancesRequest import RunInstancesRequest
from aliyunsdkecs.request.v20140526.DescribeInstancesRequest import DescribeInstancesRequest
from aliyunsdkecs.request.v20140526.DescribeNetworkInterfacesRequest import DescribeNetworkInterfacesRequest
from aliyunsdkslb.request.v20140515.DescribeLoadBalancerAttributeRequest import DescribeLoadBalancerAttributeRequest
from aliyunsdkecs.request.v20140526.RunCommandRequest import RunCommandRequest
from aliyunsdkecs.request.v20140526.SendFileRequest import SendFileRequest
from aliyunsdkecs.request.v20140526.StopInstancesRequest import StopInstancesRequest
from aliyunsdkecs.request.v20140526.DeleteInstancesRequest import DeleteInstancesRequest
from aliyunsdkecs.request.v20140526.DescribeInstanceStatusRequest import DescribeInstanceStatusRequest
from aliyunsdkcore.request import CommonRequest

logging.basicConfig(level=logging.INFO,
                    format='%(asctime)s %(filename)s[line:%(lineno)d] %(levelname)s %(message)s',
                    datefmt='%a, %d %b %Y %H:%M:%S')


def send_msg_to_feishu(webhook, key_word, msg_text):
    """发送消息到飞书"""
    headers = {'Content-Type': 'application/json'}
    payload_message = {
        "msg_type": "text",
        "content": {
            "text": '{}: {}'.format(key_word, msg_text)
        }
    }
    response = requests.request('POST', url=webhook, headers=headers, data=json.dumps(payload_message))
    logging.info(response.text)


def connect_client(access_key_id, access_key_secret, region_id):
    """
    初始化账号，连接客户端
    :param access_key_id: access key Id, type-string
    :param access_key_secret: access key secret, type-string
    :param region_id: region_id
    :return: clt
    """
    try:
        clt = AcsClient(ak=access_key_id, secret=access_key_secret, region_id=region_id)
        return clt
    except Exception as e:
        # 失败，记录报错信息，发送通知，停止并退出
        logging.error(e)
        sys.exit()


def build_create_instances_request(image_id, vswitch_id, security_group_id, zone_id, instance_type, instance_name,
                                   disk_size, disk_category, key_pair_name, tags):
    """
    购买服务器参数配置
    :param image_id: 使用的镜像信息 type-string
    :param vswitch_id: 选择的交换机 type-string
    :param security_group_id: 当前vpc类型的安全组 type-string
    :param zone_id: 服务器所在区域 type-string
    :param instance_type: 实例规格 type-string
    :param instance_name: 实例命名 type-string
    :param disk_size: 磁盘大小，单位：G，type-string
    :param disk_category: 磁盘类型 type-string
    :param key_pair_name: 密钥对名称 type-string
    :param tags: 标签 type-list, eg: [{"Key": "ecs", "Value": "rov-server.prod"}, ...]
    :return: request
    """
    request = RunInstancesRequest()
    request.set_ImageId(image_id)
    request.set_VSwitchId(vswitch_id)
    request.set_SecurityGroupId(security_group_id)
    request.set_ZoneId(zone_id)
    request.set_InstanceType(instance_type)
    request.set_InstanceName(instance_name)
    request.set_SystemDiskSize(disk_size)
    request.set_SystemDiskCategory(disk_category)
    request.set_KeyPairName(key_pair_name)
    request.set_Tags(tags)
    return request


def send_request(client, request):
    """
    发送API请求
    :param client: 客户端连接
    :param request: 请求配置
    :return: response
    """
    request.set_accept_format('json')
    try:
        response = client.do_action_with_exception(request)
        response = json.loads(response)
        # logging.info(response)
        return response
    except Exception as e:
        # 失败，记录报错信息，发送通知，停止并退出
        logging.error(e)
        sys.exit()


def check_instance_running(client, instance_ids):
    """
    检查服务器运行状态
    :param client: 客户端连接
    :param instance_ids: 实例id列表, type-list
    :return: running_count，Status为Running的实例数
    """
    try:
        request = DescribeInstancesRequest()
        request.set_InstanceIds(json.dumps(instance_ids))
        request.set_PageSize(100)
        response = send_request(client=client, request=request)
        if response.get('Code') is None:
            instances_list = response.get('Instances').get('Instance')
            running_count = 0
            running_instances = []
            for instance_detail in instances_list:
                if instance_detail.get('Status') == "Running":
                    running_count += 1
                    running_instances.append(instance_detail.get('InstanceId'))
            return running_count, running_instances
        else:
            # 失败，记录报错信息，发送通知，停止并退出
            logging.error(response)
            sys.exit()
    except Exception as e:
        # 失败，记录报错信息，发送通知，停止并退出
        logging.error(e)
        sys.exit()


def create_multiple_instances(amount, client,
                              image_id, vswitch_id, security_group_id, zone_id, instance_type, instance_name,
                              disk_size, disk_category, key_pair_name, tags):
    """
    创建多个ECS实例
    :param amount: 创建实例数 type-int 取值范围：[1, 100]
    :param client: 购买机器客户端连接
    :param image_id: 使用的镜像信息 type-string
    :param vswitch_id: 选择的交换机 type-string
    :param security_group_id: 当前vpc类型的安全组 type-string
    :param zone_id: 服务器所在区域 type-string
    :param instance_type: 实例规格 type-string
    :param instance_name: 实例命名 type-string
    :param disk_size: 磁盘大小，单位：G，type-string
    :param disk_category: 磁盘类型 type-string
    :param key_pair_name: 密钥对名称 type-string
    :param tags: 标签 type-list, eg: [{"Key": "ecs", "Value": "rov-server.prod"}, ...]
    :return:
    """
    logging.info(f"create instances start, request amount: {amount}.")
    # 1. 连接客户端
    # create_instances_clt = connect_client(
    #     access_key_id=access_key_id, access_key_secret=access_key_secret, region_id=region_id
    # )
    # 2. 请求参数配置
    request = build_create_instances_request(
        image_id=image_id, vswitch_id=vswitch_id, security_group_id=security_group_id, zone_id=zone_id,
        instance_type=instance_type, instance_name=instance_name, disk_size=disk_size, disk_category=disk_category,
        key_pair_name=key_pair_name, tags=tags
    )
    request.set_Amount(amount)
    # 3. 发送API请求，购买机器并启动
    response = send_request(client=client, request=request)
    if response.get('Code') is None:
        instance_ids = response.get('InstanceIdSets').get('InstanceIdSet')
        logging.info(f"success amount: {len(instance_ids)}, instance ids: {instance_ids}.")
        # 获取机器运行状态
        running_amount = 0
        while running_amount < amount:
            time.sleep(10)
            running_amount, running_instances = check_instance_running(client=client, instance_ids=instance_ids)
            logging.info(f"running amount: {running_amount}, running instances: {running_instances}.")
        return instance_ids
    else:
        # 失败，记录报错信息，发送通知，停止并退出
        logging.error(response)
        sys.exit()


def run_command(client, instance_ids, command):
    """
    批量执行命令
    :param client: 客户端连接
    :param instance_ids: 实例id列表, type-list
    :param command: 命令 type-string
    :return:
    """
    request = RunCommandRequest()
    request.set_accept_format('json')
    request.set_Type("RunShellScript")
    request.set_CommandContent(command)
    request.set_InstanceIds(instance_ids)
    response = send_request(client=client, request=request)
    logging.info(response)


def get_instance_ids(client, slb_id):
    """
    获取slb下所有服务器instanceId
    :param client: 客户端连接
    :param slb_id: 负载均衡id type-string
    :return: instance_ids type-list
    """
    request = DescribeLoadBalancerAttributeRequest()
    request.set_accept_format('json')
    request.set_LoadBalancerId(slb_id)
    response = send_request(client=client, request=request)
    instance_ids = [instance["ServerId"] for instance in response["BackendServers"]["BackendServer"]]
    return instance_ids


def get_ip_address(client, instance_id):
    """
    获取实例IP地址
    :param client: 客户端连接
    :param instance_id: 实例id, type-string
    :return: ip_address, type-string
    """
    request = DescribeNetworkInterfacesRequest()
    request.set_accept_format('json')
    request.set_InstanceId(instance_id)
    response = send_request(client=client, request=request)
    ip_address = response['NetworkInterfaceSets']['NetworkInterfaceSet'][0]['PrivateIpAddress']
    return ip_address


def set_weight_for_instances(client, slb_id, instance_id_list, weight):
    """
    同时设置多台服务器的slb权重，权重一样
    :param client: 客户端连接
    :param slb_id: slb_id
    :param instance_id_list: 服务器id list
    :param weight: 权重值
    :return: None
    """
    BackendServers = [{"ServerId": instance_id, "Weight": weight} for instance_id in instance_id_list]
    request = CommonRequest()
    request.set_accept_format('json')
    request.set_domain('slb.aliyuncs.com')
    request.set_version('2014-05-15')
    request.set_method('POST')
    request.set_action_name('SetBackendServers')
    request.add_query_param('BackendServers', BackendServers)
    request.add_query_param('LoadBalancerId', slb_id)
    response = send_request(client=client, request=request)
    return response


def send_file_to_ecs(client, instance_id_list, target_dir, name, content):
    """
    发送文件到ecs
    :param client:
    :param instance_id_list:
    :param target_dir: 文件存放目录 type-string
    :param name: 文件名 type-string
    :param content: 文件内容 type-string
    :return:
    """
    request = SendFileRequest()
    request.set_Content(content)
    request.set_TargetDir(target_dir)
    request.set_Name(name)
    request.set_Overwrite(True)
    request.set_InstanceIds(instance_id_list)
    response = send_request(client=client, request=request)
    return response


def stop_instances(client, instance_ids, force_stop=False):
    """
    停止实例
    :param client:
    :param instance_ids: 实例ID, type-list
    :param force_stop: 是否强制关机, True-强制关机, False-正常关机, type-bool
    :return:
    """
    request = StopInstancesRequest()
    request.set_InstanceIds(instance_ids)
    request.set_ForceStop(force_stop)
    response = send_request(client=client, request=request)
    return response


def release_instances(client, instance_ids, force=False):
    """
    释放实例
    :param client:
    :param instance_ids: instance_id, type-list
    :param force: 是否强制释放, True-强制释放, False-正常释放, type-bool
    :return:
    """
    request = DeleteInstancesRequest()
    request.set_InstanceIds(instance_ids)
    request.set_Force(force)
    response = send_request(client=client, request=request)
    return response


def get_instances_status(client, instance_ids):
    """
    获取实例运行状态
    :param client:
    :param instance_ids: instance_id, type-liist
    :return:
    """
    request = DescribeInstanceStatusRequest()
    request.set_InstanceIds(instance_ids)
    request.set_PageSize(50)
    response = send_request(client=client, request=request)
    return response


def set_instance_weight_process(client, slb_id, instance_id_list, weight_list):
    """
    修改服务器的权重值
    :param client: slb客户端连接
    :param slb_id: slb id
    :param instance_id_list: instance id list
    :param weight_list: 权重修改列表 type-list [(weight, sleep_time), ...]
    :return:
    """
    for weight, sleep_time in weight_list:
        logging.info(f"weight = {weight}")
        flag = True
        while flag:
            try:
                set_weight_for_instances(client=client, slb_id=slb_id, instance_id_list=instance_id_list, weight=weight)
                time.sleep(sleep_time)
                flag = False
            except Exception as e:
                time.sleep(10)
                continue


def add_backend_servers(client, slb_id, instances):
    """
    服务器挂载到负载均衡(必须是状态为运行中的后端服务器才可以加入负载均衡实例，每次调用最多可添加20个后端服务器)
    :param client:
    :param slb_id:
    :param instances: 实例列表 [(instance_id, ip), ...]
    :return:
    """
    try:
        request = AddBackendServersRequest()
        request.set_accept_format('json')
        request.set_LoadBalancerId(slb_id)
        backend_servers = [
            {"ServerId": instance_id, "Weight": "0", "Type": "ecs", "ServerIp": ip_address}
            for instance_id, ip_address in instances]
        request.set_BackendServers(backend_servers)
        response = client.do_action_with_exception(request)
        return response
    except Exception as e:
        logging.error(e)
        sys.exit()


def remove_backend_servers(client, slb_id, instances):
    """
    服务器从负载均衡移除(一次调用最多可以移除20个后端服务器)
    :param client:
    :param slb_id:
    :param instances: 实例列表 [(instance_id, ip), ...]
    :return:
    """
    request = RemoveBackendServersRequest()
    request.set_accept_format('json')
    request.set_LoadBalancerId(slb_id)
    backend_servers = [
        {"ServerId": instance_id, "Weight": "0", "Type": "ecs"}
        for instance_id in instances]
    request.set_BackendServers(backend_servers)
    response = client.do_action_with_exception(request)
    return response


def set_instance_weight_process_with_slbs(client, slb_id_list, instance_id_list, weight_list):
    """
    修改服务器的权重值
    :param client: slb客户端连接
    :param slb_id_list: slb id list
    :param instance_id_list: instance id list
    :param weight_list: 权重修改列表 type-list [(weight, sleep_time), ...]
    :return:
    """
    for weight, sleep_time in weight_list:
        logging.info(f"修改权重中: weight = {weight}")
        for slb_id in slb_id_list:
            flag = True
            while flag:
                try:
                    set_weight_for_instances(client=client, slb_id=slb_id, instance_id_list=instance_id_list, weight=weight)
                    logging.info(f"slb: {slb_id} finished！")
                    flag = False
                except Exception as e:
                    time.sleep(10)
                    continue
        time.sleep(sleep_time)


def add_backend_servers_with_slbs(client, slb_id_list, instances):
    """
    服务器挂载到负载均衡(必须是状态为运行中的后端服务器才可以加入负载均衡实例，每次调用最多可添加20个后端服务器)
    :param client:
    :param slb_id_list:
    :param instances: 实例列表 [(instance_id, ip), ...]
    :return:
    """
    try:
        for i in range(len(instances)//20 + 1):
            instances_list = instances[i*20:(i+1)*20]
            if len(instances_list) == 0:
                return
            for slb_id in slb_id_list:
                request = AddBackendServersRequest()
                request.set_accept_format('json')
                request.set_LoadBalancerId(slb_id)
                backend_servers = [
                    {"ServerId": instance_id, "Weight": "0", "Type": "ecs", "ServerIp": ip_address}
                    for instance_id, ip_address in instances_list]
                request.set_BackendServers(backend_servers)
                response = client.do_action_with_exception(request)
                logging.info(f"slb: {slb_id} add backend servers finished!")
            logging.info(f"i: {i}, count: {len(instances_list)}, instances: {instances_list} "
                         f"add backend servers finished!")
    except Exception as e:
        logging.error(e)
        sys.exit()


def remove_backend_servers_with_slbs(client, slb_id_list, instances):
    """
    服务器从负载均衡移除(一次调用最多可以移除20个后端服务器)
    :param client:
    :param slb_id_list:
    :param instances: 实例列表 [(instance_id, ip), ...]
    :return:
    """
    try:
        for i in range(len(instances)//20 + 1):
            instances_list = instances[i*20:(i+1)*20]
            if len(instances_list) == 0:
                return
            for slb_id in slb_id_list:
                request = RemoveBackendServersRequest()
                request.set_accept_format('json')
                request.set_LoadBalancerId(slb_id)
                backend_servers = [
                    {"ServerId": instance_id, "Weight": "0", "Type": "ecs"}
                    for instance_id in instances_list]
                request.set_BackendServers(backend_servers)
                response = client.do_action_with_exception(request)
                logging.info(f"slb: {slb_id} remove backend servers finished!")
            logging.info(f"i: {i}, count: {len(instances_list)}, instances: {instances_list} "
                         f"remove backend servers finished!")
    except Exception as e:
        logging.error(e)
        sys.exit()
